#!/usr/bin/env python3

"""
Black-Arrow file keyword searcher

Python3.5+
"""

import argparse
import os
import re
import multiprocessing as mp
import time
import subprocess
import fabulous.color as color
from faster_fifo import Queue

from typing import List, Union


RETYPE = type(
    re.compile("a")
)  # since re module apparently doesn't have good compiled types
EDITOR = os.environ.get("EDITOR", "nvim")  # Default editor
DEVMODE = False
QUEUE_SIZE = 1_000_000_000_000


def start_search(args: argparse.Namespace):
    """
    This function is separated out in order to use code as module
    """
    global DEVMODE
    DEVMODE = args.dev

    start_time = time.time()

    args.ignore = args.ignore + ["build", "\.min\.js", ".git"]

    # compile for performance reasons
    try:
        ignore_re = re.compile("a^")
        if args.ignore:
            ignore_re = re.compile("|".join(args.ignore))
        filename_re = re.compile(".*")
        if args.filename:
            filename_re = re.compile("|".join(args.filename))

        if args.lower or args.regex != args.regex.lower():
            search_re = re.compile(args.regex)
        else:
            search_re = re.compile(args.regex, flags=re.IGNORECASE)
    except re.error as e:
        print(color.red("Error, bad regular expression:"))
        raise

    # get numworkers
    # default to numcores
    # backup is 6 or specified by user
    try:
        numworkers = mp.cpu_count()
    except NotImplementedError:
        numworkers = args.workers or 6

    mp.set_start_method('fork')

    search_queue = Queue(QUEUE_SIZE)
    output = Queue(QUEUE_SIZE)
    final_queue = Queue(QUEUE_SIZE)  # Use final queue for external output
    processes = []

    indexer = mp.Process(
        name="indexer",
        target=index_worker,
        args=(
            args.directories,
            ignore_re,
            filename_re,
            numworkers,
            search_queue,
            args.depth,
        ),
    )
    indexer.start()
    processes.append(indexer)

    for i in range(numworkers):
        worker = mp.Process(
            name="worker-{}".format(i + 1),
            target=file_searching_worker,
            args=(i, search_re, args.replace, search_queue, output),
        )
        worker.start()
        processes.append(worker)
    printer = mp.Process(
        name="printer",
        target=print_worker,
        args=(start_time, numworkers, output, final_queue, args.pipe, args.edit),
    )
    printer.start()
    processes.append(printer)

    return processes, final_queue


def index_worker(
    directories: List[str],
    ignore_re: RETYPE,
    filename_re: RETYPE,
    workers: int,
    search_queue: Queue,
    depth: int,
    block=False,
) -> None:
    for dir in list(set(directories)):  # no duplicates
        for subdir, folders, files in os.walk(dir):
            # if depth exceeds the required depth, do not walk deeper
            if depth is not None and subdir.count(os.sep) >= depth:
                del folders[:]

            for question_file in files:
                should_we_search = filename_re.search(question_file) is not None
                do_we_ignore = ignore_re.search(question_file) is None
                if should_we_search and do_we_ignore:
                    # we don't want to block, this process should be fastest
                    search_queue.put(
                        subdir + "/" + question_file, block=True, timeout=10
                    )  # faster than os.path.join
    for i in range(workers):
        search_queue.put("EXIT")  # poison pill workers


def file_searching_worker(
    worker_num: int, regex: RETYPE, replace: Union[str, None], search_queue: Queue, output: Queue
) -> None:
    line_count = 0
    file_count = 0
    found_count = 0
    # https://stackoverflow.com/questions/566746/how-to-get-console-window-width-in-python
    rows, columns = os.popen("stty size", "r").read().split()
    # Max width of printed line is 3/4 column count
    maxwidth = int(3 * int(columns) / 4)
    while True:
        # we want to block this thread until we get search_queue
        name = search_queue.get()
        if name == "EXIT":
            output.put(("EXIT" + str(worker_num), line_count, file_count, found_count))
            break

        file_count += 1
        try:
            new_text = None
            with open(name, "r") as ofile:
                flag = []
                i = 1
                for line in ofile:
                    if regex.search(line):
                        found_string = line.split("\n")[0]
                        if len(found_string) > maxwidth:
                            found_string = found_string[:maxwidth] + "..."
                        flag.append((i, found_string))
                    i += 1
                line_count += i
                if flag:
                    found_count += 1
                    for value in flag:
                        if replace is not None:
                            output.put((name, value[0], value[1], regex, replace))
                        else:
                            output.put((name, value[0], value[1], regex))
                if replace is not None:
                    ofile.seek(0)  # reset to beginning
                    new_text = regex.subn(replace, ofile.read())[0]
            if replace is not None:
                with open(name, "w") as ofile:
                    ofile.write(new_text)
        except:
            pass


def print_worker(
    start_time: float,
    worker_count: int,
    output: Queue,
    final_queue: Queue,
    pipemode: bool,
    editmode: bool,
) -> None:
    file_count = 0
    found_count = 0
    exit_count = 0
    line_count = 0
    file_list = []
    while True:
        statement = output.get()
        if statement[0][:4] == "EXIT":
            exit_count += 1
            line_count += statement[1]
            file_count += statement[2]
            found_count += statement[3]

            if DEVMODE:
                print(statement[0])
                print(exit_count, worker_count)

            if exit_count == worker_count:
                break

        else:
            if len(statement) == 4:
                filename, linenum, matched, line = statement
                replace = None
            else:
                filename, linenum, matched, line, replace = statement
            # final_queue.put(filename, linenum)
            if not DEVMODE:
                if pipemode:
                    print("{}	{}	{}".format(filename, linenum, matched))
                else:
                    print(
                        "{}:{}\n\t{}".format(
                            filename,
                            color.fg256("#00ff00", linenum),
                            insert_colour(matched, line, extra_str=replace),
                        )
                    )

                    file_list.append((statement[1], statement[0]))

    final_queue.put("EXIT")

    if not pipemode:
        print(
            (
                "---------------\n"
                "Files Searched: {:,}\n"
                "Files Matched: {:,}\n"
                "Lines Searched: {:,}\n"
                "Duration: {:.3f}"
            ).format(file_count, found_count, line_count, time.time() - start_time)
        )

    if editmode:
        files_to_edit = ["+{} {}".format(num, name) for num, name in file_list]
        call_args = [_ for _ in files_to_edit[0].split(" ")]
        orientation = 0
        for f in files_to_edit[1:]:
            call_args.append('+"{} {}"'.format("sp" if orientation else "vsp", f))
            orientation ^= 1
        if len(call_args) > 10:
            print(color.red("Cowardly only accepting the first 10 files for editing"))
        call_string = "{} {}".format(EDITOR, " ".join(call_args[:10]))
        subprocess.call(call_string, shell=True)


def insert_colour(matchstring: str, regex: RETYPE, extra_str=None) -> str:
    """
    Given some string and a regex, color the match inside that string

    :param matchstring:
    :param regex:
    :param extra_str:
    :return:
    """
    if extra_str is None:
        replace_str = str(color.fg256("yellow", r"\g<0>"))
    else:
        replace_str = str(color.fg256("yellow", r"(\g<0> -> {})".format(extra_str)))
    return re.sub("^[ \t]+", "", re.sub(regex, replace_str, matchstring))



import argparse
from typing import Optional

from . import blackarrow as ba


def main():
    args = get_args()
    processes, final_queue = ba.start_search(args)
    try:
        # This doesn't work as it is hanging on that queue for memory safety
        # print_process.join()  # Wait main thread until printer is done

        # need to instead listen to that queue and treat it as poison - like the others
        while True:
            poison = final_queue.get()
            if poison == 'EXIT':
                break

    except (KeyboardInterrupt, EOFError):  # kill all on ctrl+c/d
        [p.terminate() for p in processes]


def get_args(manual_args: Optional[str] = None) -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    regex_group = parser.add_mutually_exclusive_group(required=True)
    regex_group.add_argument(
        "regex_positional",
        metavar="R",
        type=str,
        default=None,
        nargs="?",
        help="Search term (regular expression)",
    )
    parser.add_argument(
        "-d",
        "--directories",
        type=str,
        default=["."],
        nargs="+",
        help="Director(y|ies) to run against",
    )
    parser.add_argument(
        "-i",
        "--ignore",
        type=str,
        default=[],
        nargs="+",
        help="Things to ignore (regular expressions)",
    )
    parser.add_argument(
        "-f",
        "--filename",
        type=str,
        default=[],
        nargs="+",
        help="Filename search term(s)",
    )
    parser.add_argument(
        "-w",
        "--workers",
        type=int,
        default=None,
        help=("Number of workers to use (default numcores, with fallback 6 unless set)"),
    )
    parser.add_argument(
        "-p",
        "--pipe",
        action="store_true",
        default=False,
        help=('Run in "pipe" mode with brief output'),
    )
    parser.add_argument(
        "-e", "--edit", action="store_true", default=False, help=("Edit the files?")
    )
    parser.add_argument(
        "-l",
        "--lower",
        action="store_true",
        default=False,
        help=("Check strict lower case?"),
    )
    parser.add_argument(
        "-r",
        "--replace",
        type=str,
        default=None,
        help="Replace text found in place with supplied",
    )
    parser.add_argument(
        "-D",
        "--depth",
        type=int,
        default=None,
        required=False,
        help="Directory depth to search in",
    )
    parser.add_argument(
        "--dev",
        action="store_true",
        default=False,
        help=("Run in development mode (NO OUTPUT)"),
    )

    if manual_args is not None:
        args = parser.parse_args(args=manual_args)
    else:
        args = parser.parse_args()
    args.regex = args.regex_positional
    return args



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['blackarrow', 'blackarrow.test']

package_data = \
{'': ['*']}

install_requires = \
['fabulous>=0.4.0,<0.5.0', 'faster-fifo>=1.4.5,<2.0.0']

entry_points = \
{'console_scripts': ['ba = blackarrow:main', 'blackarrow = blackarrow:main']}

setup_kwargs = {
    'name': 'blackarrow',
    'version': '1.1.4',
    'description': 'A fast keyword searcher',
    'long_description': '"Universal" Text Finder\n=======================\n\n[![Test Status](https://travis-ci.org/TheDataLeek/black-arrow.svg?branch=master)](https://travis-ci.org/TheDataLeek/black-arrow)\n[![Coverage Status](https://coveralls.io/repos/github/TheDataLeek/black-arrow/badge.svg?branch=master)](https://coveralls.io/github/TheDataLeek/black-arrow?branch=master)\n\nIt\'s basically just grep in python... Nothing fancy, just an easy extensible way\nto find things....\n\nYeah, I know, it\'s "reinventing the wheel" but ehhh, this is easier to extend to\ncover any and all weird cases without having to memorize a bunch of obscure\ncombinations of bash commands.\n\nThat out of the way, let\'s talk about what this *actually does*...\n\n## Quick Overview\n\nEssentially Black Arrow (or `ba` for short) is a way to search through every line in every file and\nfind matching keyword *or* regular expressions. It uses smart case, so if the search term is all\nlowercase it defaults to case-insensitive (mostly for ease of use).\n\n![png](./img/demo.png)\n\nYou can also supply it with regular expressions and it handles them natively.\n\n![png](./img/demo2.png)\n\nOther features include -\n\n* excluding certain files or paths,\n* replacing any matches that have been found,\n* specifying the max depth to search (good for large directory structures),\n* "pipe" mode for unix piping,\n* open the files in the `$EDITOR` for manual parsing. \n\n## Installation\n\n```\n┬─[zoe@fillory:~/Dropbox/Projects/black-arrow]─[09:22:12 PM]\n╰─>$ pip install --user blackarrow\n```\n\n## Black-Arrow Script\n\n```bash\n┬─[zoe@fillory:~/Dropbox/Projects/black-arrow]─[09:33:40 PM]\n╰─>$ ./black-arrow/blackarrow.py -h\nusage: ba [-h] [-d DIRECTORIES [DIRECTORIES ...]] [-i IGNORE [IGNORE ...]]\n          [-f FILENAME [FILENAME ...]] [-w WORKERS] [-p] [-e] [-l]\n          [-r REPLACE] [-D DEPTH] [--dev]\n          R\n\npositional arguments:\n  R                     Search term (regular expression)\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -d DIRECTORIES [DIRECTORIES ...], --directories DIRECTORIES [DIRECTORIES ...]\n                        Director(y|ies) to run against\n  -i IGNORE [IGNORE ...], --ignore IGNORE [IGNORE ...]\n                        Things to ignore (regular expressions)\n  -f FILENAME [FILENAME ...], --filename FILENAME [FILENAME ...]\n                        Filename search term(s)\n  -w WORKERS, --workers WORKERS\n                        Number of workers to use (default numcores, with\n                        fallback 6 unless set)\n  -p, --pipe            Run in "pipe" mode with brief output\n  -e, --edit            Edit the files?\n  -l, --lower           Check strict lower case?\n  -r REPLACE, --replace REPLACE\n                        Replace text found in place with supplied\n  -D DEPTH, --depth DEPTH\n                        Directory depth to search in\n  --dev                 Run in development mode (NO OUTPUT)\n```\n\n#### The Name\n\n*"Arrow! Black arrow! I have saved you to the last. You have never failed me and\nI have always recovered you. I had you from my father and he from of old. If\never you came from the forges of the true king under the Mountain, go now and\nspeed well!"*\n\n― J.R.R. Tolkien, The Hobbit\n',
    'author': 'Zoe Farmer',
    'author_email': 'zoe@dataleek.io',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/TheDataLeek/black-arrow',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



from ..blackarrow import index_worker

import re
import pytest
import queue


@pytest.fixture
def results():
    """
    def index_worker(
        directories: List[str],
        ignore_re: RETYPE,
        filename_re: RETYPE,
        workers: int,
        search_queue: mp.Queue,
        depth: int,
        block=False,
    ) -> None:
    :return:
    """
    input_queue = queue.Queue()
    index_worker(
        ['sample'],  # directories to look in
        re.compile("a^"), # things to ignore
        re.compile('test'),  # things to search for
        1,  # num workers
        input_queue,  # where to put results
        None,   # depth (not needed)
        block=True   # whether or not to block
    )

    vals = []
    while not input_queue.empty():
        vals.append(input_queue.get())
    return vals


def test_not_empty(results):
    assert len(results) != 0


def test_correct_results(results):
    assert (
        len(
            set(results).difference(
                {'sample/tester.txt',
                 'sample/tester2.txt',
                 'EXIT'}
            )
        ) == 0
    )
